==> ubee512-1.4.0.tar.gz: ubee512-1.4.0/src/paknet.c <==
//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                               paknet module                                *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// PAK and NET selection.  This is used for selecting PAK and NET ROMs when
// emulating a ROM model.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.4.0 - 2 October 2007, S.J.Kay
// - Fully implemented PAK and NET ROMs selection for various ROM models.
//
// v1.0.0 - 3 July 2007, S.J.Kay
// - Created a new file and implement the PAK and NET ports

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "paknet.h"
#include "ubee512.h"
#include "vdu.h"
#include "z80.h"
#include "mz80.h"
#include "memmap.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
static char romimage[512];

UINT8 pakdata;
static int paknumb=-1;			// current PAK ROM
static int paksize;			// size of current PAK ROM
static char pakx[0x4000];		// 2 x 8K banks of PAK code

static char netx[0x4000];		// 4 x 4K banks of NET code

extern int model;
extern int rom_model;
extern char userhome_romspath[];

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Load a PAK ROM image from file.
//
// Each model uses a different file name for each PAK ROM so that the user
// can link to the actual ROM file concerned.
//
//   pass: pak 			PAK ROM number
// return: void
//==============================================================================
static void loadpak (int pak)
{
 int pak_mask;
 
 char *PAK_PPC85[] =
   {
    "PREM-C.ROM",	// Wordbee
    "PREM-E.ROM",	// Basic help
    "PREM-F.ROM",	// Busycalc/PC85 Menu
    "PREM-G.ROM",	// Graphics/Database
    "PREMVTEX.ROM",	// Vidotex
    "PREM-I.ROM",	// Shell
    "PREM-PAK6.ROM",	// User PAK 6
    "PREM-PAK7.ROM"	// User PAK 7
   };

 char *PAK_PC85[] =
   {
    "PC85_PAK0.ROM",	// User PAK 0
    "PC85_PAK1.ROM",	// User PAK 1
    "PC85_PAK2.ROM",	// User PAK 2
    "PC85_PAK3.ROM",	// User PAK 3
    "PC85_PAK4.ROM",	// User PAK 4
    "PC85_PAK5.ROM",	// User PAK 5
    "PC85_PAK6.ROM",	// User PAK 6
    "PC85_PAK7.ROM"	// User PAK 7
   };

   
 char *PAK_PC[] =
   {
    "PC_EDASM.ROM",	// Editor Assembler
    "PC_WBEE.ROM",	// Wordbee
    "PC_PAK2.ROM",	// User PAK 2
    "PC_PAK3.ROM",	// User PAK 3
    "PC_PAK4.ROM",	// User PAK 4
    "PC_PAK5.ROM",	// User PAK 5
    "PC_PAK6.ROM",	// User PAK 6
    "PC_PAK7.ROM"	// User PAK 7
   };

 char *PAK_IC[] =
   {
    "IC_WBEE.ROM",	// Wordbee
    "IC_PAK1.ROM",	// User PAK 1    
    "IC_PAK2.ROM",	// User PAK 2
    "IC_PAK3.ROM",	// User PAK 3
    "IC_PAK4.ROM",	// User PAK 4
    "IC_PAK5.ROM",	// User PAK 5
    "IC_PAK6.ROM",	// User PAK 6
    "IC_PAK7.ROM"	// User PAK 7
   };

 char *PAK_2MHZ[] =
   {
    "2MHZ_PAK0.ROM",	// User PAK 0
   };

 romimage[0] = 0;
 strcat(romimage, userhome_romspath); 

 pakdata = pak;

 switch (model)
    {
     case MOD_PPC85 : pak_mask = B8(00000111);
                      strcat(romimage, PAK_PPC85[pakdata & pak_mask]);
                      break;
     case  MOD_PC85 : pak_mask = B8(00000111);
                      strcat(romimage, PAK_PC85[pakdata & pak_mask]);
                      break;
     case    MOD_PC : pak_mask = B8(00000111);
                      strcat(romimage, PAK_PC[pakdata & pak_mask]);
                      break;                     
     case    MOD_IC : pak_mask = B8(00000111);
                      strcat(romimage, PAK_IC[pakdata & pak_mask]);
                      break;
     case  MOD_2MHZ : pak_mask = B8(00000000);
                      strcat(romimage, PAK_2MHZ[0]);
                      break;
            default : return;
    }            

 if (paknumb != (pakdata & pak_mask))
    {
     paknumb = pakdata & pak_mask;
     paksize = memmap_loadrom(romimage, pakx, sizeof(pakx));
    }

// if PAK ROM is 16K and PC85 or later then bit 3 determines lower/upper 8K in use (A13)
 if ((paksize >= 0x4000) && (pakdata & B8(00001000)) && ((model == MOD_PPC85) || (model == MOD_PC85)))
    memcpy(z80mem + PAK_ADDR, pakx + 0x2000, 0x2000);
 else   
    memcpy(z80mem + PAK_ADDR, pakx, 0x2000);
} 

//==============================================================================
// Load a NET ROM image from file.
//
// Each model uses a different file name so that the user can link to the
// actual ROM file concerned.
//
//   pass: void
// return: void
//==============================================================================
static void loadnet (void)
{
 romimage[0] = 0;
 strcat(romimage, userhome_romspath); 

 switch (model)
    {
     case MOD_PPC85 : strcat(romimage, "PREM_NETWORK.ROM");
                      break;
     case  MOD_PC85 : strcat(romimage, "PC85_NETWORK.ROM");
                      break;
     case    MOD_PC : strcat(romimage, "PC_NETWORK.ROM");
                      break;
     case    MOD_IC : strcat(romimage, "IC_NETWORK.ROM");
                      break;
     case  MOD_2MHZ : strcat(romimage, "2MHZ_NETWORK.ROM");
                      break;
            default : return;          
    }
    
 memmap_loadrom(romimage, netx, sizeof(netx));
 memcpy(z80mem + NET_ADDR, netx, 0x1000);
}

//==============================================================================
// PAK and NET Initialise.
//
// Load the PAK0 and NET ROM into memory
//
//   pass: void
// return: int				0
//==============================================================================
int paknet_init (void)
{
 if (rom_model)
    {
// load default PAK ROM
     loadpak(0);

// load NET ROM
     loadnet();
    } 
 return 0;
}

//==============================================================================
// PAK and NET  de-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int paknet_deinit (void)
{
 if (rom_model)
    {
    }
 return 0;
}

//==============================================================================
// PAK and NET  reset
//
// A reset causes bank 0 of network ROM to be selected.  This should work for
// all ROM based models.
//
//   pass: void
// return: int				0
//==============================================================================
int paknet_reset (void)
{
 if (rom_model)
    {
     memcpy(z80mem + NET_ADDR, netx, 0x1000);
    } 
 return 0;
}

//==============================================================================
// NET read - Port function
//
// Load the NET (x) ROM to 0xE000-0xEFFF
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0xFFFF
//==============================================================================
UINT16 paknet_nsel_r (UINT16 port, struct z80PortRead *port_s)
{
 struct mz80context z80;

 uint8_t bank;
 uint8_t bank_mask;
 uint8_t opcode;

 if (rom_model)
    {
     switch (rom_model)
        {
         case MOD_PPC85 : 
         case  MOD_PC85 : bank_mask = B8(00000011);
                          break;
         case    MOD_PC : bank_mask = B8(00000001);
                          break;
         case    MOD_IC : 
         case  MOD_2MHZ : bank_mask = B8(00000000);
                          break;
        }             
     
     mz80GetContext(&z80);    

     opcode = (uint8_t)z80mem[z80.z80pc - 2];     
     if (opcode == 0xdb)

// MZ80 has endian problem, work around for now
//        bank = z80.z80af.half.a & bank_mask;
        bank = (int)z80.z80af.af & bank_mask;
     else
//        bank = z80.z80bc.half.b & bank_mask;
        bank = ((int)z80.z80bc.bc >> 8) & bank_mask;        

     memcpy(z80mem + NET_ADDR, netx + (0x1000 * bank), 0x1000);
     
#if 1
     if (opcode == 0xdb)
        {
         printf("paknet_nsel_r: Z80 Instruction=0x%02x  Reg A=0x%02x\n", opcode, bank);
         printf("paknet_nsel_r: z80af=0x%08x\n\n", z80.z80af.af);
        }
     else
        {
         printf("paknet_nsel_r: Z80 Instruction=0x%02x  Reg B=0x%02x\n", opcode, bank);
         printf("paknet_nsel_r: z80bc=0x%08x\n\n", z80.z80bc.bc);
        }        
        
#endif
     if (debuglvl > 0)
        fprintf(debuglog, "paknet_nsel_r: Z80 Instruction=0x%02x\n  bank=%d\n", opcode, bank);
    }
 return 0xFFFF;
}

//==============================================================================
// PAK write - Port function
//
// Load the PAK (x) ROM to 0xC000-0xDFFF
//
//   pass: UINT16 port
//         UINT8 data			PAK ROM # (0-255)
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void paknet_psel_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (rom_model)
    {
#if 0
     printf("paknet_psel_w: port=0x%02x data=0x%02x\n", port, data);
#endif
     if (debuglvl > 0)
        fprintf(debuglog, "paknet_psel_w: port=0x%02x data=0x%02x\n", port, data);

     loadpak(data);
    }
}

==> ubee512-2.1.0.tar.gz: ubee512-2.1.0/src/paknet.c <==
//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                               paknet module                                *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// PAK and NET selection.  This is used for selecting PAK and NET ROMs when
// emulating a ROM model.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.1.0 - 25 October 2007, S.J.Kay
// - Attempt a repair of a known bad Telcom v3.2.1 ROM image if detected.
// - Fixed bug in paknet_nsel_r() function. Switch statement had used incorrect
//   variable (rom_model) instead of (model) so value would have been always 1
//   and no matching case.  New code uses (modelx.model)
//
// v1.4.0 - 2 October 2007, S.J.Kay
// - Fully implemented PAK and NET ROMs selection for various ROM models.
//
// v1.0.0 - 3 July 2007, S.J.Kay
// - Created a new file and implement the PAK and NET ports

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "paknet.h"
#include "ubee512.h"
#include "vdu.h"
#include "z80.h"
#include "mz80.h"
#include "memmap.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
static char romimage[512];

UINT8 pakdata;
static int paknumb=-1;			// current PAK ROM
static int paksize;			// size of current PAK ROM
static uint8_t pakx[0x4000];		// 2 x 8K banks of PAK code

static uint8_t netx[0x4000];		// 4 x 4K banks of NET code

extern int model;
extern model_t modelx;

extern char userhome_romspath[];

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Load a PAK ROM image from file.
//
// Each model uses a different file name for each PAK ROM so that the user
// can link to the actual ROM file concerned.
//
//   pass: pak 			PAK ROM number
// return: void
//==============================================================================
static void paknet_loadpak (int pak)
{
 int pak_mask;
 
 char *PAK_PPC85[] =
   {
    "PREM-C.ROM",	// Wordbee
    "PREM-E.ROM",	// Basic help
    "PREM-F.ROM",	// Busycalc/PC85 Menu
    "PREM-G.ROM",	// Graphics/Database
    "PREMVTEX.ROM",	// Vidotex
    "PREM-I.ROM",	// Shell
    "PREM-PAK6.ROM",	// User PAK 6
    "PREM-PAK7.ROM"	// User PAK 7
   };

 char *PAK_PC85[] =
   {
    "PC85_PAK0.ROM",	// User PAK 0
    "PC85_PAK1.ROM",	// User PAK 1
    "PC85_PAK2.ROM",	// User PAK 2
    "PC85_PAK3.ROM",	// User PAK 3
    "PC85_PAK4.ROM",	// User PAK 4
    "PC85_PAK5.ROM",	// User PAK 5
    "PC85_PAK6.ROM",	// User PAK 6
    "PC85_PAK7.ROM"	// User PAK 7
   };

 char *PAK_PC[] =
   {
    "PC_EDASM.ROM",	// Editor Assembler
    "PC_WBEE.ROM",	// Wordbee
    "PC_PAK2.ROM",	// User PAK 2
    "PC_PAK3.ROM",	// User PAK 3
    "PC_PAK4.ROM",	// User PAK 4
    "PC_PAK5.ROM",	// User PAK 5
    "PC_PAK6.ROM",	// User PAK 6
    "PC_PAK7.ROM"	// User PAK 7
   };

 char *PAK_IC[] =
   {
    "IC_WBEE.ROM",	// Wordbee
    "IC_PAK1.ROM",	// User PAK 1    
    "IC_PAK2.ROM",	// User PAK 2
    "IC_PAK3.ROM",	// User PAK 3
    "IC_PAK4.ROM",	// User PAK 4
    "IC_PAK5.ROM",	// User PAK 5
    "IC_PAK6.ROM",	// User PAK 6
    "IC_PAK7.ROM"	// User PAK 7
   };

 char *PAK_2MHZ[] =
   {
    "2MHZ_PAK0.ROM",	// User PAK 0
   };

 romimage[0] = 0;
 strcat(romimage, userhome_romspath); 

 pakdata = pak;

 switch (model)
    {
     case MOD_PPC85 : pak_mask = B8(00000111);
                      strcat(romimage, PAK_PPC85[pakdata & pak_mask]);
                      break;
     case  MOD_PC85 : pak_mask = B8(00000111);
                      strcat(romimage, PAK_PC85[pakdata & pak_mask]);
                      break;
     case    MOD_PC : pak_mask = B8(00000111);
                      strcat(romimage, PAK_PC[pakdata & pak_mask]);
                      break;                     
     case    MOD_IC : pak_mask = B8(00000111);
                      strcat(romimage, PAK_IC[pakdata & pak_mask]);
                      break;
     case  MOD_2MHZ : pak_mask = B8(00000000);
                      strcat(romimage, PAK_2MHZ[0]);
                      break;
            default : return;
    }            

 if (paknumb != (pakdata & pak_mask))
    {
     paknumb = pakdata & pak_mask;
     paksize = memmap_loadrom(romimage, (char*)pakx, sizeof(pakx));
    }

// if PAK ROM is 16K and PC85 or later then bit 3 determines lower/upper 8K in use (A13)
 if ((paksize >= 0x4000) && (pakdata & B8(00001000)) && ((model == MOD_PPC85) || (model == MOD_PC85)))
    memcpy(z80mem + PAK_ADDR, pakx + 0x2000, 0x2000);
 else   
    memcpy(z80mem + PAK_ADDR, pakx, 0x2000);
} 

//==============================================================================
// Load a NET ROM image from file.
//
// Each model uses a different file name so that the user can link to the
// actual ROM file concerned.
//
// A check is made for known problem ROM(s) and patched,  the data used is the
// output of the cmp command in Linux.  Two ROMs are compared,  the address is
// decimal and the data values are octal.
//
// cmp -l badrom.rom goodrom.rom
//
//   pass: void
// return: void
//==============================================================================
static void paknet_loadnet (void)
{
 const romfix_t romfix_telc321[] =
 {
// OFS+1  BAD  GOOD  
  {38,   0322, 0332},
  {39,   0347, 0357},
  {432,  0141, 0351},
  {1408, 0145, 0345},
  {1504, 047,  0347},
  {2080, 0155, 0355},
  {2132, 0240, 0341},
  {2297, 0176, 0376},
  {2334, 0166, 0367},
  {2444, 0101, 0315},
  {2665, 0216, 0256},
  {2678, 00,   0346},
  {3002, 0122, 0377},
  {3084, 044,  0344},
  {3260, 020,  0335},
  {3269, 067,  0267},
  {3290, 042,  0356},
  {3296, 057,  0157},
  {3324, 031,  0335},
  {3559, 04,   0137},
  {3849, 055,  0355},
  {3916, 0330, 0370},		// this is the last of the check data
  {-1,   0,    0   },
  {3969, 0304, 0104},		// fixes the missing 'D' in 'iv / power ^'
  {-1,   0,    0   }            // and new line in the calculator.
 };

 int i;
 
 romimage[0] = 0;
 strcat(romimage, userhome_romspath); 

 switch (model)
    {
     case MOD_PPC85 : strcat(romimage, "PREM_NETWORK.ROM");
                      break;
     case  MOD_PC85 : strcat(romimage, "PC85_NETWORK.ROM");
                      break;
     case    MOD_PC : strcat(romimage, "PC_NETWORK.ROM");
                      break;
     case    MOD_IC : strcat(romimage, "IC_NETWORK.ROM");
                      break;
     case  MOD_2MHZ : strcat(romimage, "2MHZ_NETWORK.ROM");
                      break;
            default : return;          
    }
    
 memmap_loadrom(romimage, (char*)netx, sizeof(netx));

// Check if this is a known bad Telcom v3.2.1 ROM image and repair it.
// It has to detect all known bad values to be repaired.
 i = 0;
 while (romfix_telc321[i].ofs != -1)
    {
     if (netx[romfix_telc321[i].ofs - 1] != romfix_telc321[i].bad)
        break;
     i++;
    }

// we get to the end of the list only if all the values in the list checked were bad
 if (romfix_telc321[i].ofs == -1)
    {
     printf("paknet_loadnet: Bad version of Telcom v3.2.1 detected,  attempting repair\n");
     i = 0;
     while (romfix_telc321[i].ofs != -1)
        {
         netx[romfix_telc321[i].ofs - 1] = romfix_telc321[i].good;
         i++;
        }
// do some other repair stuff on that was not part of the check test.
     i++;
     while (romfix_telc321[i].ofs != -1)
        {
         netx[romfix_telc321[i].ofs - 1] = romfix_telc321[i].good;
         i++;
        }        
    }

// Copy the 4K or part of a 8/16K ROM to the NET ROM address in Z80 memory
 memcpy(z80mem + NET_ADDR, netx, 0x1000);
}

//==============================================================================
// PAK and NET Initialise.
//
// Load the PAK0 and NET ROM into memory
//
//   pass: void
// return: int				0
//==============================================================================
int paknet_init (void)
{
 if (modelx.rom)
    {
// load default PAK ROM
     paknet_loadpak(0);

// load NET ROM
     paknet_loadnet();
    } 
 return 0;
}

//==============================================================================
// PAK and NET  de-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int paknet_deinit (void)
{
 if (modelx.rom)
    {
    }
 return 0;
}

//==============================================================================
// PAK and NET  reset
//
// A reset causes bank 0 of network ROM to be selected.  This should work for
// all ROM based models.
//
//   pass: void
// return: int				0
//==============================================================================
int paknet_reset (void)
{
 if (modelx.rom)
    {
     memcpy(z80mem + NET_ADDR, netx, 0x1000);
    } 
 return 0;
}

//==============================================================================
// NET read - Port function
//
// Load the NET (x) ROM to 0xE000-0xEFFF
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0xFFFF
//==============================================================================
UINT16 paknet_nsel_r (UINT16 port, struct z80PortRead *port_s)
{
 struct mz80context z80;

 uint8_t bank;
 uint8_t bank_mask;
 uint8_t opcode;

 if (modelx.rom)
    {
     switch (model)
        {
         case MOD_PPC85 : 
         case  MOD_PC85 : bank_mask = B8(00000011);
                          break;
         case    MOD_PC : bank_mask = B8(00000001);
                          break;
         case    MOD_IC : 
         case  MOD_2MHZ : bank_mask = B8(00000000);
                          break;
                default : bank_mask = 0;
                          break;
        }             
     
     mz80GetContext(&z80);    

     opcode = (uint8_t)z80mem[z80.z80pc - 2];     
     if (opcode == 0xdb)

// MZ80 has endian problem, work around for now
//        bank = z80.z80af.half.a & bank_mask;
        bank = (int)z80.z80af.af & bank_mask;
     else
//        bank = z80.z80bc.half.b & bank_mask;
        bank = ((int)z80.z80bc.bc >> 8) & bank_mask;        

     memcpy(z80mem + NET_ADDR, netx + (0x1000 * bank), 0x1000);
     
#if 0
     if (opcode == 0xdb)
        {
         printf("paknet_nsel_r: Z80 Instruction=0x%02x  Reg A=0x%02x\n", opcode, bank);
         printf("paknet_nsel_r: z80af=0x%08x\n\n", z80.z80af.af);
        }
     else
        {
         printf("paknet_nsel_r: Z80 Instruction=0x%02x  Reg B=0x%02x\n", opcode, bank);
         printf("paknet_nsel_r: z80bc=0x%08x\n\n", z80.z80bc.bc);
        }        
        
#endif
     if (debuglvl > 0)
        fprintf(debuglog, "paknet_nsel_r: Z80 Instruction=0x%02x\n  bank=%d\n", opcode, bank);
    }
 return 0xFFFF;
}

//==============================================================================
// PAK write - Port function
//
// Load the PAK (x) ROM to 0xC000-0xDFFF
//
//   pass: UINT16 port
//         UINT8 data			PAK ROM # (0-255)
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void paknet_psel_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    {
#if 0
     printf("paknet_psel_w: port=0x%02x data=0x%02x\n", port, data);
#endif
     if (debuglvl > 0)
        fprintf(debuglog, "paknet_psel_w: port=0x%02x data=0x%02x\n", port, data);

     paknet_loadpak(data);
    }
}
